import asyncio
import concurrent.futures
from typing import Any

from google.api_core.exceptions import NotFound, PreconditionFailed
from google.cloud import storage  # type: ignore

from smartjob.app.exception import SmartJobException
from smartjob.app.storage import StoragePort


class GcsStorageAdapter(StoragePort):
    def __init__(self, max_workers: int = 10):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        self.storage_client = storage.Client()

    def sync_upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        if not destination_path:
            raise SmartJobException("destination_path is required")
        if not destination_bucket:
            raise SmartJobException("bucket is required")
        b = self.storage_client.bucket(destination_bucket)
        blob = b.blob(destination_path)
        kwargs: dict[str, Any] = {}
        if only_if_not_exists:
            kwargs["if_generation_match"] = 0
        try:
            blob.upload_from_string(content, **kwargs, retry=None)
        except PreconditionFailed:
            pass

    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        cf_future = self.executor.submit(
            self.sync_upload,
            content,
            destination_bucket,
            destination_path,
            only_if_not_exists=only_if_not_exists,
        )
        await asyncio.wrap_future(cf_future)

    def sync_copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        kwargs: dict[str, Any] = {"retry": None}
        if only_if_not_exists:
            kwargs["if_generation_match"] = 0
        sb = self.storage_client.bucket(source_bucket)
        source_blob = sb.blob(source_path)
        db = self.storage_client.bucket(destination_bucket)
        sb.copy_blob(source_blob, db, destination_path, **kwargs)

    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        cf_future = self.executor.submit(
            self.sync_copy,
            source_bucket,
            source_path,
            destination_bucket,
            destination_path,
            only_if_not_exists=only_if_not_exists,
        )
        await asyncio.wrap_future(cf_future)

    def sync_download(
        self,
        source_bucket: str,
        source_path: str,
    ) -> bytes:
        b = self.storage_client.bucket(source_bucket)
        blob = b.blob(source_path)
        try:
            return blob.download_as_bytes(retry=None)
        except NotFound:
            return b""

    async def download(
        self,
        source_bucket: str,
        source_path: str,
    ) -> bytes:
        cf_future = self.executor.submit(
            self.sync_download,
            source_bucket,
            source_path,
        )
        return await asyncio.wrap_future(cf_future)



import asyncio
import io
import tarfile
from dataclasses import dataclass, field

import docker
from stlog import getLogger

from smartjob.app.storage import StoragePort

logger = getLogger("smartjob.storage.docker")


# inspired from https://stackoverflow.com/questions/46390309/how-to-copy-a-file-from-host-to-container-using-docker-py-docker-sdk
def copy_to_container(
    container: docker.models.containers.Container, content: str | bytes, dst: str
):
    tar_stream = io.BytesIO()
    src_stream = io.BytesIO()
    if isinstance(content, str):
        length = src_stream.write(content.encode())
    else:
        length = src_stream.write(content)
    src_stream.seek(0)
    with tarfile.open(fileobj=tar_stream, mode="w|") as tar:
        info = tarfile.TarInfo(name=dst)
        info.size = length
        tar.addfile(info, fileobj=src_stream)
    container.put_archive("/", tar_stream.getvalue())


def read_from_container(
    container: docker.models.containers.Container, src: str
) -> bytes:
    try:
        bits, _ = container.get_archive(src)
    except Exception:
        return b""
    res = io.BytesIO()
    for chunk in bits:
        res.write(chunk)
    res.seek(0)
    with tarfile.open(fileobj=res, mode="r|") as tar:
        for member in tar:
            memberfile = tar.extractfile(member)
            if memberfile:
                return memberfile.read()
    return b""


@dataclass
class DockerStorageAdapter(StoragePort):
    docker_client: docker.DockerClient = field(default_factory=docker.DockerClient)
    mount_points_cache: dict[str, str] = field(default_factory=dict, init=False)
    lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False)

    async def get_or_create_dummy_container_for_uploading_to_volumes(
        self,
    ) -> docker.models.containers.Container:
        # see https://github.com/moby/moby/issues/25245
        async with self.lock:
            staging_volume = self.get_or_create_staging_docker_volume()
            try:
                res = self.docker_client.containers.get("smartjob-dummy-container")
                logger.debug("smartjob-dummy-container found => let's reuse it")
                return res
            except Exception:
                pass
            logger.debug("let's create smartjob-dummy-container")
            try:
                self.docker_client.images.get("docker.io/alpine:latest")
            except Exception:
                logger.debug("let's pull docker.io/alpine:latest image")
                self.docker_client.images.pull("docker.io/alpine:latest")
            return self.docker_client.containers.create(
                image="docker.io/alpine:latest",
                name="smartjob-dummy-container",
                volumes={staging_volume.name: {"bind": "/staging", "mode": "rw"}},
            )

    def get_or_create_staging_docker_volume(self) -> docker.models.volumes.Volume:
        for volume in self.docker_client.volumes.list():
            if volume.name == "smartjob-staging":
                logger.debug("found smartjob-staging volume => let's reuse it")
                return volume
        # we have to create it
        logger.debug("let's create the smartjob-staging volume")
        return self.docker_client.volumes.create(name="smartjob-staging")

    async def download(self, source_bucket: str, source_path: str) -> bytes:
        dummy = await self.get_or_create_dummy_container_for_uploading_to_volumes()
        return read_from_container(dummy, "/staging/" + source_path)

    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        # Note: only_if_not_exists is not supported!
        if len(content) == 0 and destination_path.endswith("/"):
            # this is a special case for bucket-like storage
            # let's just create an empty file
            destination_path = destination_path + ".empty"
        dummy = await self.get_or_create_dummy_container_for_uploading_to_volumes()
        copy_to_container(dummy, content, "/staging/" + destination_path)

    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        raise Exception("not supported for the moment")



from abc import abstractmethod
from dataclasses import dataclass, field

from stlog import getLogger

from smartjob.app.execution import Execution
from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.storage import StorageService
from smartjob.infra.adapters.storage.gcs import GcsStorageAdapter

logger = getLogger("smartjob.executor.gcp")


class GCPExecutionResultFuture(ExecutionResultFuture):
    pass


@dataclass
class GCPExecutor(ExecutorPort):
    """Abstract base class for GCP executors."""

    max_workers: int = 10
    _storage_service: StorageService | None = field(default=None, init=False)

    def __post_init__(self):
        self._storage_service = StorageService(
            adapter=GcsStorageAdapter(max_workers=self.max_workers)
        )

    @property
    def storage_service(self) -> StorageService:
        assert self._storage_service is not None
        return self._storage_service

    @abstractmethod
    def staging_mount_path(self, execution: Execution) -> str:
        pass

    def get_storage_service(self) -> StorageService:
        return self.storage_service



import asyncio
import json
from dataclasses import dataclass, field

from smartjob.app.storage import StoragePort


@dataclass
class DummyStorageAdapter(StoragePort):
    sleep: float = 1.0
    keep_in_memory: bool = False
    data: dict[str, bytes | str] = field(default_factory=dict, init=False)

    async def _sleep(self):
        await asyncio.sleep(self.sleep)

    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        await self._sleep()
        if self.keep_in_memory:
            self.data[destination_path] = content

    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        await self._sleep()
        if self.keep_in_memory:
            self.data[destination_path] = (
                f"copied from {source_bucket}/{source_path}".encode()
            )

    async def download(
        self,
        source_bucket: str,
        source_path: str,
    ) -> bytes:
        await self._sleep()
        return json.dumps({"dummy": "result"}).encode()



from dataclasses import dataclass


@dataclass
class RetryConfig:
    """Retry configuration."""

    max_attempts: int = 3
    """FIXME"""

    max_attempts_schedule: int | None = None
    """FIXME"""

    max_attempts_execute: int | None = None
    """FIXME"""

    max_attempts_download: int | None = None
    """FIXME"""

    @property
    def _max_attempts_schedule(self) -> int:
        return self.max_attempts_schedule or self.max_attempts

    @property
    def _max_attempts_execute(self) -> int:
        return self.max_attempts_schedule or self.max_attempts

    @property
    def _max_attempts_download(self) -> int:
        return self.max_attempts_download or self.max_attempts

    def __post_init__(self):
        for field in (
            "max_attempts",
            "max_attempts_schedule",
            "max_attempts_execute",
            "max_attempts_download",
        ):
            if getattr(self, field) is not None and getattr(self, field) < 1:
                raise ValueError(f"{field} must be >= 1")



import asyncio
import concurrent.futures
import logging
import sys
from dataclasses import dataclass, field
from threading import Lock

from google.cloud import aiplatform
from google.cloud.aiplatform_v1.types import custom_job as custom_job_v1
from google.cloud.aiplatform_v1.types import job_state as gca_job_state
from google.cloud.aiplatform_v1.types.env_var import EnvVar
from google.cloud.aiplatform_v1.types.machine_resources import DiskSpec, MachineSpec
from stlog import LogContext, getLogger

from smartjob.app.execution import Execution, ExecutionResult
from smartjob.app.executor import ExecutionResultFuture
from smartjob.infra.adapters.executor.gcp import GCPExecutionResultFuture, GCPExecutor

aiplatform_mutex = Lock()
aiplatform_initialized: bool = False
logger = getLogger("smartjob.executor.vertex")


def monkey_patch_aiplatform_jobs_logger():
    from google.cloud.aiplatform.jobs import _LOGGER

    for handler in _LOGGER.handlers:
        _LOGGER.removeHandler(handler)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.WARNING)
    _LOGGER.addHandler(handler)
    logging.getLogger("google.cloud.aiplatform.jobs").setLevel(logging.WARNING)


def init_aiplatform():
    global aiplatform_initialized
    with aiplatform_mutex:
        if not aiplatform_initialized:
            aiplatform.init()
            monkey_patch_aiplatform_jobs_logger()
            aiplatform_initialized = True


class VertexCustomJob(aiplatform.CustomJob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_created = asyncio.Event()

    @property
    def id(self) -> str:
        return self.resource_name.split("/")[-1]

    async def get_log_url(self, project) -> str:
        await self.resource_created.wait()
        return self._get_log_url(project)

    def _get_log_url(self, project) -> str:
        try:
            return f"https://console.cloud.google.com/logs/query;query=resource.labels.job_id%3D%22{self.id}%22?project={project}"
        except Exception:
            return "https://no-log-url.com/sorry"

    @property
    def success(self) -> bool:
        return self.state == gca_job_state.JobState.JOB_STATE_SUCCEEDED


class VertexExecutionResultFuture(GCPExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        return future.result()


@dataclass
class VertexExecutorAdapter(GCPExecutor):
    _executor: concurrent.futures.ThreadPoolExecutor | None = field(
        default=None, init=False
    )

    def __post_init__(self):
        super().__post_init__()
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.max_workers
        )

    @property
    def executor(self) -> concurrent.futures.ThreadPoolExecutor:
        assert self._executor is not None
        return self._executor

    def init_aiplatform_if_needed(self):
        init_aiplatform()

    def sync_run(
        self, custom_job: VertexCustomJob, execution: Execution, log_context: dict
    ) -> ExecutionResult:
        self.init_aiplatform_if_needed()
        job = execution.job
        with LogContext.bind(
            **log_context
        ):  # we need to rebind here as we are now in another thread
            logger.info(
                "Let's trigger a new execution of Vertex Job...",
                docker_image=job.docker_image,
                overridden_args=execution.add_envs_as_string,
                add_envs=execution.overridden_args_as_string,
            )
            try:
                custom_job.submit(
                    disable_retries=True,
                    timeout=execution.config._timeout_config.timeout_seconds,
                    service_account=execution.config.service_account,
                )
                custom_job.resource_created.set()
                logger.debug("Resource created")
                custom_job._block_until_complete()
            except RuntimeError:
                pass
            finally:
                custom_job.resource_created.set()
            return ExecutionResult._from_execution(
                execution,
                custom_job.success,
                custom_job._get_log_url(execution.config._project),
            )

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        job = execution.job
        config = execution.config
        loop = asyncio.get_event_loop()
        vertex_name = f"{job.full_name}-{execution.id}"
        with LogContext.bind(
            vertex_name=vertex_name, project=config._project, region=config._region
        ):
            custom_job = VertexCustomJob(
                display_name=vertex_name,
                labels=execution.labels,
                project=execution.config._project,
                location=execution.config._region,
                staging_bucket=execution.config._staging_bucket,
                worker_pool_specs=[
                    custom_job_v1.WorkerPoolSpec(
                        replica_count=1,
                        container_spec=custom_job_v1.ContainerSpec(
                            image_uri=job.docker_image,
                            args=execution.overridden_args,
                            env=[
                                EnvVar(name=k, value=v)
                                for k, v in execution.add_envs.items()
                            ],
                        ),
                        machine_spec=MachineSpec(
                            machine_type=execution.config._machine_type,
                            accelerator_type=execution.config._accelerator_type,
                            accelerator_count=execution.config._accelerator_count,
                        ),
                        disk_spec=DiskSpec(
                            boot_disk_type=execution.config._boot_disk_type,
                            boot_disk_size_gb=execution.config._boot_disk_size_gb,
                        ),
                    )
                ],
            )
            future = loop.run_in_executor(
                self.executor, self.sync_run, custom_job, execution, LogContext.getall()
            )
            log_url = await custom_job.get_log_url(execution.config._project)
            return VertexExecutionResultFuture(
                asyncio.ensure_future(future),
                execution=execution,
                storage_service=self.storage_service,
                gcs_output_path=f"{config._staging_bucket}/{execution.output_relative_path}",
                log_url=log_url,
            )

    def staging_mount_path(self, execution: Execution) -> str:
        return f"/gcs/{execution.config._staging_bucket_name}"

    def get_name(self) -> str:
        return "vertex"



import asyncio
from dataclasses import dataclass, field

from smartjob.app.execution import Execution, ExecutionResult
from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.storage import StorageService
from smartjob.infra.adapters.storage.dummy import DummyStorageAdapter


class DummyExecutionResultFuture(ExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        return future.result()

    async def _get_output(self) -> dict | list | str | int | float | bool | None:
        return None


def make_storage_service() -> StorageService:
    return StorageService(adapter=DummyStorageAdapter())


@dataclass
class DummyExecutorAdapter(ExecutorPort):
    sleep: float = 1.0
    storage_service: StorageService = field(default_factory=make_storage_service)

    async def work(self, execution: Execution) -> ExecutionResult:
        await asyncio.sleep(self.sleep)
        return ExecutionResult._from_execution(
            execution, True, "https://no-log-url.com/sorry"
        )

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        return DummyExecutionResultFuture(
            asyncio.create_task(self.work(execution)),
            execution,
            storage_service=self.storage_service,
            log_url="https://no-log-url.com/sorry",
        )

    def get_name(self):
        return "dummy"

    def get_input_path(self, execution: Execution) -> str:
        return "/dummy"

    def get_output_path(self, execution: Execution) -> str:
        return "/dummy"

    def get_storage_service(self) -> StorageService:
        return self.storage_service

    def staging_mount_path(self, execution: Execution) -> str:
        return "/empty"



from abc import ABC, abstractmethod
from dataclasses import dataclass


class StoragePort(ABC):
    @abstractmethod
    async def download(self, source_bucket: str, source_path: str) -> bytes:
        pass

    @abstractmethod
    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        pass

    @abstractmethod
    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ) -> str:
        pass


@dataclass
class StorageService:
    adapter: StoragePort

    async def download(self, source_bucket: str, source_path: str) -> bytes:
        return await self.adapter.download(source_bucket, source_path)

    async def upload(
        self,
        content: bytes | str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        return await self.adapter.upload(
            content, destination_bucket, destination_path, only_if_not_exists
        )

    async def copy(
        self,
        source_bucket: str,
        source_path: str,
        destination_bucket: str,
        destination_path: str,
        only_if_not_exists: bool = True,
    ):
        return await self.adapter.copy(
            source_bucket,
            source_path,
            destination_bucket,
            destination_path,
            only_if_not_exists=only_if_not_exists,
        )



from smartjob.app.executor import ExecutorPort, ExecutorService

# Default max workers for vertex executor and for file uploader (cloud run executor is not multi-threaded)
DEFAULT_MAX_WORKERS = 10
__cache: dict[str, ExecutorService] = {}


def get_executor_service(
    type: str = "cloudrun",
    max_workers: int = DEFAULT_MAX_WORKERS,
    use_cache: bool = True,
) -> ExecutorService:
    """Return a singleton instance of ExecutorService (initialized on first call with given arguments).

    Args:
        type: The type of executor to use ('cloudrun', 'vertex' or 'docker').
        max_workers: Maximum number of workers for the vertex executor and for the file uploader.
        use_cache: if set to True, the same instance will be returned for the same type.

    Returns:
        Instance of ExecutorService.

    """
    adapter: ExecutorPort
    if type == "cloudrun":
        from smartjob.infra.adapters.executor.cloudrun import CloudRunExecutorAdapter

        adapter = CloudRunExecutorAdapter(max_workers=max_workers)
    elif type == "vertex":
        from smartjob.infra.adapters.executor.vertex import VertexExecutorAdapter

        adapter = VertexExecutorAdapter(max_workers=max_workers)
    elif type == "docker":
        from smartjob.infra.adapters.executor.docker import DockerExecutorAdapter

        adapter = DockerExecutorAdapter(max_workers=max_workers)
    else:
        raise ValueError(
            f"Invalid executor type: {type} => must be 'cloudrun', 'vertex' or 'docker'"
        )
    if not use_cache or type not in __cache:
        __cache[type] = ExecutorService(
            adapter=adapter,
        )
    return __cache[type]



import datetime
import json
import typing
from dataclasses import dataclass

from smartjob.app.exception import SmartJobException
from smartjob.app.storage import StorageService


@dataclass
class Input:
    """Abstract base class for inputs. Do not use directly."""

    filename: str
    """File name to be used in the input path (without /)."""

    async def _create(self, bucket: str, path: str, storage_service: StorageService):
        raise NotImplementedError("must be implemented in subclasses")

    def __post_init__(self):
        if "/" in self.filename:
            raise SmartJobException("filename cannot contain /")


@dataclass
class BytesInput(Input):
    """Represents an input given as bytes or str."""

    content: bytes | str
    """Content to be written to the input path (can be bytes or string)."""

    async def _create(self, bucket: str, path: str, storage_service: StorageService):
        await storage_service.upload(self.content, bucket, f"{path}/{self.filename}")


@dataclass
class JsonInput(Input):
    """Represents an input given as a JSON-serializable object."""

    content: typing.Any
    """Content to be seriliazed to JSON and written to the input path."""

    def _json_default(self, value):
        if isinstance(value, datetime.datetime):
            return value.isoformat()
        return "*** NOT JSON SERIALIZABLE ***"

    def _as_bytes(self) -> bytes:
        return json.dumps(self.content, indent=4, default=self._json_default).encode(
            "utf-8"
        )

    async def _create(self, bucket: str, path: str, storage_service: StorageService):
        bytesInput = BytesInput(filename=self.filename, content=self._as_bytes())
        await bytesInput._create(bucket, path, storage_service)


@dataclass
class LocalPathInput(Input):
    """Represents an input given as a local file path."""

    local_path: str
    """Local path to the file to be copied to the input path."""

    async def _create(self, bucket: str, path: str, storage_service: StorageService):
        with open(self.local_path, "rb") as f:
            content = f.read()
        bytesInput = BytesInput(filename=self.filename, content=content)
        await bytesInput._create(bucket, path, storage_service)


@dataclass
class GcsInput(Input):
    """Can be used to copy a file from a GCS bucket to the input path."""

    gcs_path: str
    """GCS path to the file to be copied (e.g. gs://my-bucket/my-file.txt)."""

    def __post_init__(self):
        super().__post_init__()
        if not self.gcs_path.startswith("gs://"):
            raise SmartJobException("gcs_path must start with gs://")
        if "/" not in self.gcs_path[5:]:
            raise SmartJobException(
                "gcs_path must contain a / (excluding the gs:// prefix)"
            )

    async def _create(self, bucket: str, path: str, storage_service: StorageService):
        source_bucket = self.gcs_path[5:].split("/")[0]
        source_path = self.gcs_path[5:].split("/", 1)[1]
        await storage_service.copy(
            source_bucket, source_path, bucket, f"{path}/{self.filename}"
        )



from dataclasses import dataclass, field

# DEFAULT_NAMESPACE is the default namespace for jobs.
DEFAULT_NAMESPACE = "default"


@dataclass
class SmartJob:
    """Base class for smart jobs."""

    name: str
    """Name of the job."""

    docker_image: str
    """Docker image to use for the job.

    Example: `docker.io/python:3.12`.
    """

    namespace: str = DEFAULT_NAMESPACE
    """Namespace of the job."""

    add_envs: dict[str, str] = field(default_factory=dict)
    """Environnement variables to add in the container.

    Note: `INPUT_PATH` and `OUTPUT_PATH` will be automatically injected (if relevant).
    """

    overridden_args: list[str] = field(default_factory=list)
    """Container arguments (including command) to use.

    Notes:
        - if not set, the default container image arguments will be used.
        - the placeholder {{INPUT}} will be automatically replaced by the local full path of the input directory.

    """

    python_script_path: str = ""
    """Local path to a python script to execute in the container.

    Note: if set, it will override overridden_args."""

    @property
    def full_name(self) -> str:
        """Return the full name of the job (namespace + name)."""
        return f"{self.namespace}-{self.name}"



from smartjob.app.exception import SmartJobException, SmartJobTimeoutException
from smartjob.app.execution import ExecutionConfig, ExecutionResult
from smartjob.app.executor import ExecutionResultFuture, ExecutorService
from smartjob.app.input import BytesInput, GcsInput, Input, JsonInput, LocalPathInput
from smartjob.app.job import SmartJob
from smartjob.app.retry import RetryConfig
from smartjob.app.timeout import TimeoutConfig
from smartjob.infra.controllers.lib import get_executor_service

__all__ = [
    "get_executor_service",
    "SmartJobException",
    "SmartJobTimeoutException",
    "SmartJob",
    "ExecutorService",
    "ExecutionResult",
    "ExecutionResultFuture",
    "Input",
    "JsonInput",
    "BytesInput",
    "GcsInput",
    "LocalPathInput",
    "RetryConfig",
    "TimeoutConfig",
    "ExecutionConfig",
]



import asyncio
import concurrent.futures
from dataclasses import dataclass, field

import docker
from stlog import LogContext, getLogger

from smartjob.app.execution import Execution, ExecutionResult
from smartjob.app.executor import ExecutionResultFuture, ExecutorPort
from smartjob.app.storage import StorageService
from smartjob.infra.adapters.storage.docker import DockerStorageAdapter

logger = getLogger("smartjob.executor.docker")


class DockerExecutionResultFuture(ExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        return future.result()


def make_storage_service() -> StorageService:
    return StorageService(adapter=DockerStorageAdapter())


@dataclass
class DockerExecutorAdapter(ExecutorPort):
    max_workers: int = 10
    sleep: float = 1.0
    storage_service: StorageService = field(default_factory=make_storage_service)
    _executor: concurrent.futures.ThreadPoolExecutor | None = field(
        default=None, init=False
    )

    def __post_init__(self):
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.max_workers
        )

    @property
    def executor(self) -> concurrent.futures.ThreadPoolExecutor:
        assert self._executor is not None
        return self._executor

    def sync_run(
        self, execution: Execution, container_id: str, log_context: dict
    ) -> ExecutionResult:
        docker_client = docker.DockerClient()
        with LogContext.bind(
            **log_context
        ):  # we need to rebind here as we are now in another thread
            container = docker_client.containers.get(container_id)
            res = container.wait()
            logger.debug("Container stopped")
            return ExecutionResult._from_execution(
                execution, res["StatusCode"] == 0, f"docker logs -f {container.id}"
            )

    def load_docker_image_if_needed(self, docker_image: str):
        docker_client = docker.DockerClient()
        try:
            docker_client.images.get(docker_image)
            logger.debug(f"{docker_image} image already exists => let's use it")
        except Exception:
            logger.debug(f"let's pull {docker_image} image...")
            docker_client.images.pull(docker_image)
            logger.debug(f"{docker_image} image pulled")

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        docker_client = docker.DockerClient()
        loop = asyncio.get_event_loop()
        job = execution.job
        name = f"{job.name}-{execution.id}"
        await loop.run_in_executor(
            self.executor, self.load_docker_image_if_needed, job.docker_image
        )
        container = docker_client.containers.create(
            name=name,
            image=job.docker_image,
            command=execution.overridden_args,
            auto_remove=False,
            volumes={"smartjob-staging": {"bind": "/staging", "mode": "rw"}},
            environment=execution.add_envs,
        )
        with LogContext.bind(container_id=container.id):
            logger.info(
                "Container created",
                container_name=name,
                image=job.docker_image,
                command=execution.overridden_args_as_string,
                env=execution.add_envs_as_string,
            )
            container.start()
            logger.info("Container started")
            future = loop.run_in_executor(
                self.executor,
                self.sync_run,
                execution,
                container.id,
                LogContext.getall(),
            )
            return DockerExecutionResultFuture(
                asyncio.ensure_future(future),
                execution,
                storage_service=self.storage_service,
                log_url=f"docker logs -f {container.id}",
            )

    def get_name(self):
        return "docker"

    def get_storage_service(self) -> StorageService:
        return self.storage_service

    def staging_mount_path(self, execution: Execution) -> str:
        return "/staging"



import datetime
import json
import os
import shlex
from dataclasses import dataclass, field

from stlog import getLogger

from smartjob.app.exception import SmartJobException
from smartjob.app.input import Input
from smartjob.app.job import SmartJob
from smartjob.app.retry import RetryConfig
from smartjob.app.timeout import TimeoutConfig
from smartjob.app.utils import unique_id

DEFAULT_MACHINE_TYPE = "n1-standard-4"
DEFAULT_ACCELERATOR_COUNT = 0
DEFAULT_ACCELERATOR_TYPE = "ACCELERATOR_TYPE_UNSPECIFIED"
DEFAULT_BOOT_DISK_TYPE = "pd-ssd"
DEFAULT_BOOT_DISK_SIZE_GB = 100
DEFAULT_CPU = 1.0
DEFAULT_MEMORY_GB = 0.5

logger = getLogger("smartjob.execution")


def read_from_env(name: str, default: str | None = None) -> str | None:
    value = os.environ.get(name)
    if value is None:
        return default
    logger.debug("Read %s=%s from env", name, value)
    return value


@dataclass
class ExecutionConfig:
    """ExecutionConfig is the configuration for a job execution."""

    retry_config: RetryConfig | None = None
    """Retry configuration."""

    timeout_config: TimeoutConfig | None = None
    """Timeout configuration."""

    project: str | None = None
    """GCP project (only for cloudrun/vertex executor)."""

    region: str | None = None
    """GCP region (only for cloudrun/vertex executor)."""

    labels: dict[str, str] | None = None
    """GCP labels (only for cloudrun/vertex executor)."""

    staging_bucket: str | None = None
    """Staging bucket (for input, output and/or loading python_script_path, only for cloudrun/vertex executor)."""

    service_account: str | None = None
    """Service account (email) to use for the job execution (only for cloudrun/vertex executor)."""

    cpu: float | None = None
    """Number of requested CPUs (only for cloudrun executor)."""

    memory_gb: float | None = None
    """Memory in Gb (only for cloudrun executor)."""

    machine_type: str | None = None
    """Machine type (only for vertex executor)."""

    accelerator_type: str | None = None
    """Accelerator type (only for vertex executor)."""

    accelerator_count: int | None = None
    """Number of accelerators (only for vertex executor)."""

    boot_disk_type: str | None = None
    """Boot disk type (only for vertex executor)."""

    boot_disk_size_gb: int | None = None
    """Boot disk size in Gb (only for vertex executor)."""

    vpc_connector_network: str | None = None
    """VPC connector network (only for cloudrun executor)."""

    vpc_connector_subnetwork: str | None = None
    """VPC connector subnetwork (only for cloudrun executor)."""

    @property
    def _retry_config(self) -> RetryConfig:
        assert self.retry_config is not None
        return self.retry_config

    @property
    def _labels(self) -> dict[str, str]:
        assert self.labels is not None
        return self.labels

    @property
    def _timeout_config(self) -> TimeoutConfig:
        assert self.timeout_config is not None
        return self.timeout_config

    @property
    def _staging_bucket(self) -> str:
        if self.staging_bucket is None:
            return "dummy"
        return self.staging_bucket

    @property
    def _project(self) -> str:
        assert self.project is not None
        return self.project

    @property
    def _region(self) -> str:
        assert self.region is not None
        return self.region

    @property
    def _staging_bucket_name(self) -> str:
        if self.staging_bucket is None:
            return "dummy"
        if self.staging_bucket.startswith("gs://"):
            return self.staging_bucket[5:]
        return self.staging_bucket

    @property
    def _cpu(self) -> float:
        assert self.cpu is not None
        return self.cpu

    @property
    def _memory_gb(self) -> float:
        assert self.memory_gb is not None
        return self.memory_gb

    @property
    def _machine_type(self) -> str:
        assert self.machine_type is not None
        return self.machine_type

    @property
    def _accelerator_type(self) -> str:
        assert self.accelerator_type is not None
        return self.accelerator_type

    @property
    def _accelerator_count(self) -> int:
        assert self.accelerator_count is not None
        return self.accelerator_count

    @property
    def _boot_disk_type(self) -> str:
        assert self.boot_disk_type is not None
        return self.boot_disk_type

    @property
    def _boot_disk_size_gb(self) -> int:
        assert self.boot_disk_size_gb is not None
        return self.boot_disk_size_gb

    def fix_timeout_config(self) -> None:
        if self.timeout_config is None:
            self.timeout_config = TimeoutConfig()

    def fix_for_executor_name(self, executor_name: str) -> None:
        # Prechecks
        if executor_name == "cloudrun":
            for field_name in [
                "machine_type",
                "accelerator_type",
                "accelerator_count",
                "boot_disk_type",
                "boot_disk_size_gb",
            ]:
                if getattr(self, field_name) is not None:
                    logger.warning(f"{field_name} is ignored for cloudrun executor")
        elif executor_name == "vertex":
            for field_name in [
                "cpu",
                "memory_gb",
                "vpc_connector_network",
                "vpc_connector_subnetwork",
            ]:
                if getattr(self, field_name) is not None:
                    logger.warning(f"{field_name} is ignored for vertex executor")
        elif executor_name == "docker":
            for field_name in [
                "machine_type",
                "accelerator_type",
                "accelerator_count",
                "boot_disk_type",
                "boot_disk_size_gb",
                "cpu",
                "memory_gb",
                "labels",
                "vpc_connector_network",
                "vpc_connector_subnetwork",
            ]:
                if getattr(self, field_name) is not None:
                    logger.warning(f"{field_name} is ignored for docker executor")
            self.staging_bucket = "smartjob-staging"  # we force this special value
            if self.staging_bucket is not None:
                logger.warning(
                    "staging_bucket is not supported for docker executor => let's ignore it"
                )
        if executor_name in ("vertex", "docker"):
            if self.retry_config is not None:
                if self.retry_config._max_attempts_execute != 1:
                    logger.warning(
                        f"retry_config.max_attempts_execute if not supported for {executor_name} executor => let's change the setting to 1"
                    )
                    self.retry_config.max_attempts_execute = 1
            else:
                self.retry_config = RetryConfig(max_attempts_execute=1)
        # Default values
        self.fix_timeout_config()
        if self.retry_config is None:
            self.retry_config = RetryConfig()
        if self.staging_bucket is None:
            self.staging_bucket = read_from_env("SMARTJOB_STAGING_BUCKET")
        if self.service_account is None:
            self.service_account = read_from_env("SMARTJOB_SERVICE_ACCOUNT")
        if self.project is None:
            self.project = read_from_env("SMARTJOB_PROJECT")
        if self.region is None:
            self.region = read_from_env("SMARTJOB_REGION")
        if self.machine_type is None:
            self.machine_type = DEFAULT_MACHINE_TYPE
        if self.accelerator_count is None:
            self.accelerator_count = DEFAULT_ACCELERATOR_COUNT
        if self.accelerator_type is None:
            self.accelerator_type = DEFAULT_ACCELERATOR_TYPE
        if self.boot_disk_type is None:
            self.boot_disk_type = DEFAULT_BOOT_DISK_TYPE
        if self.boot_disk_size_gb is None:
            self.boot_disk_size_gb = DEFAULT_BOOT_DISK_SIZE_GB
        if self.cpu is None:
            self.cpu = DEFAULT_CPU
        if self.memory_gb is None:
            self.memory_gb = DEFAULT_MEMORY_GB
        if self.labels is None:
            self.labels = {}
        # Post checks
        if executor_name in ("cloudrun", "vertex"):
            for field_name in ["staging_bucket", "project", "region"]:
                if getattr(self, field_name) is None:
                    raise SmartJobException(
                        f"{field_name} is required for cloudrun executor"
                    )
            if not self.staging_bucket.startswith("gs://"):  # type: ignore
                raise SmartJobException("staging_bucket must start with gs://")


@dataclass
class Execution:
    job: SmartJob
    config: ExecutionConfig
    id: str = field(default_factory=unique_id)
    inputs: list[Input] = field(default_factory=list)
    created: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc),
        init=False,
    )
    cancelled: bool = field(default=False, init=False)
    add_envs: dict[str, str] = field(default_factory=dict)
    overridden_args: list[str] = field(default_factory=list)

    @property
    def base_dir(self) -> str:
        dat = self.created.strftime("%Y-%m-%d")
        tim = self.created.strftime("%H%M")
        return f"{self.job.full_name}/{dat}/{tim}/{self.id}"

    @property
    def input_relative_path(self) -> str:
        return f"{self.base_dir}/input"

    @property
    def output_relative_path(self) -> str:
        return f"{self.base_dir}/output"

    @property
    def add_envs_as_string(self) -> str:
        return ", ".join([f"{x}={y}" for x, y in self.add_envs.items()])

    @property
    def overridden_args_as_string(self) -> str:
        return shlex.join(self.overridden_args)

    @property
    def labels(self) -> dict[str, str]:
        return self.config._labels | {
            "smartjob": "true",
            "smartjob.namespace": self.job.namespace,
        }


@dataclass
class ExecutionResult:
    """ExecutionResult is the (final) result of a job execution.

    Attributes:
        success: Whether the job has succeeded or not.
        created: The datetime when the job has started.
        stopped: The datetime when the job has stopped.
        execution_id: The execution id of the job.
        job_name: The name of the job.
        job_namespace: The namespace of the job.
        log_url: The execution log url.
        full_input_path: The full input path (starting with gs://).
        full_output_path: The full output path (starting with gs//).
        json_output: if the job has created a json file named smartjob.json in the output directory, it will be stored/decoded here.

    """

    success: bool
    created: datetime.datetime
    stopped: datetime.datetime
    execution_id: str
    job_name: str
    job_namespace: str
    log_url: str
    json_output: dict | list | str | float | int | bool | None = None

    def __bool__(self) -> bool:
        return self.success or False

    def __str__(self) -> str:
        if self.success:
            state = "SUCCESS"
        else:
            state = "FAILURE"
        if self.json_output is not None:
            json_output = json.dumps(self.json_output, indent=4)
        else:
            json_output = "None"
        res = f"""ExecutionResult(
    job_name={self.job_name}, job_namespace={self.job_namespace},
    execution_id={self.execution_id},
    state={state}, duration_seconds={self.duration_seconds},
    json_output={json_output}
)"""
        return res

    @property
    def duration_seconds(self) -> int:
        """The duration of the job in seconds."""
        return (self.stopped - self.created).seconds

    @classmethod
    def _from_execution(
        cls,
        execution: Execution,
        success: bool,
        log_url: str,
    ) -> "ExecutionResult":
        """Create a ExecutionResult from a SmartJobExecution."""
        return cls(
            created=execution.created,
            execution_id=execution.id,
            job_name=execution.job.name,
            job_namespace=execution.job.namespace,
            success=success,
            stopped=datetime.datetime.now(tz=datetime.timezone.utc),
            log_url=log_url,
        )



import asyncio
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Coroutine

from stlog import LogContext, getLogger
from tenacity import AsyncRetrying, RetryError, stop_after_attempt, wait_exponential

from smartjob.app.exception import SmartJobTimeoutException
from smartjob.app.execution import Execution, ExecutionConfig, ExecutionResult
from smartjob.app.input import Input
from smartjob.app.job import SmartJob
from smartjob.app.storage import StorageService

logger = getLogger("smartjob.executor")


class ExecutionResultFuture(ABC):
    """Future-like object to get the result of a job execution.

    Attributes:
        log_url: The execution log url.

    """

    def __init__(
        self,
        task: asyncio.Task,
        execution: Execution,
        storage_service: StorageService,
        log_url: str,
        **kwargs,
    ):
        self._execution: Execution = execution
        self.__result: ExecutionResult | None = None
        self.__task = task
        self.__task.add_done_callback(self._task_done)
        self.__done_callback_called = asyncio.Event()
        self._storage_service = storage_service
        self.log_url = log_url
        self._log_context: dict[str, Any] = {}

    def _cancel(self):
        if not self.__result and not self._execution.cancelled:
            self._execution.cancelled = True
            self.__task.cancel()

    async def _get_output(self) -> dict | list | str | int | float | bool | None:
        with LogContext.bind(**self._log_context):
            staging_bucket_name = self._execution.config._staging_bucket_name
            staging_bucket = self._execution.config._staging_bucket
            smartjob_path = f"{self._execution.output_relative_path}/smartjob.json"
            logger.info(
                "Downloading smartjob.json from output (if exists)...",
                path=f"{staging_bucket}/{smartjob_path}",
            )
            raw = await self._storage_service.download(
                staging_bucket_name, smartjob_path
            )
            if raw == b"":
                logger.debug("No smartjob.json found in output")
                return None
            try:
                res = json.loads(raw)
                logger.debug("smartjob.json downloaded/decoded")
                return res
            except Exception:
                logger.warning("smartjob.json is not a valid json")
                return None

    async def _get_output_with_timeout_and_retries(self):
        config = self._execution.config
        try:
            async with asyncio.timeout(
                config._timeout_config._timeout_seconds_download
            ):
                try:
                    async for attempt in AsyncRetrying(
                        reraise=True,
                        wait=wait_exponential(min=1, max=300, multiplier=1),
                        stop=stop_after_attempt(
                            config._retry_config._max_attempts_download
                        ),
                    ):
                        with attempt:
                            with LogContext.bind(
                                attempt=attempt.retry_state.attempt_number
                            ):
                                return await self._get_output()
                except RetryError:
                    raise
                raise Exception("Unreachable code")
        except asyncio.TimeoutError:
            raise SmartJobTimeoutException(
                "Timeout reached while downloading ({config._timeout_config._timeout_seconds_download} seconds)"
            )

    async def result(self) -> ExecutionResult:
        """Wait and return the result of the job execution.

        This is a coroutine. You have to wait for it to complete with (for example):
        `res = await future.result()`

        Returns:
            The result of the job execution.

        """
        await self.__done_callback_called.wait()
        assert self.__result is not None
        if self.__result.success:
            self.__result.json_output = (
                await self._get_output_with_timeout_and_retries()
            )
        return self.__result

    def done(self) -> bool:
        """Return True if the job execution is done."""
        return self.__task.done()

    @property
    def execution_id(self) -> str:
        "The execution unique identifier."
        return self._execution.id

    @abstractmethod
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        pass

    def _task_done(self, future: asyncio.Future):
        with LogContext.bind(**self._log_context):
            self.__result = self._get_result_from_future(future)
            if not self._execution.cancelled:
                if self.__result.success:
                    logger.info(
                        "Smartjob execution succeeded",
                        log_url=self.log_url,
                        duration_seconds=self.__result.duration_seconds,
                    )
                else:
                    logger.warn(
                        "Smartjob execution failed",
                        log_url=self.log_url,
                        duration_seconds=self.__result.duration_seconds,
                    )
            self.__done_callback_called.set()


class ExecutorPort(ABC):
    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        pass

    @abstractmethod
    def get_storage_service(self) -> StorageService:
        pass

    @abstractmethod
    def staging_mount_path(self, execution: Execution) -> str:
        pass


@dataclass
class ExecutorService:
    adapter: ExecutorPort
    _storage_service: StorageService | None = field(default=None, init=False)
    executor_name: str = field(default="", init=False)

    @property
    def storage_service(self) -> StorageService:
        assert self._storage_service is not None
        return self._storage_service

    def __post_init__(self):
        self.executor_name = self.adapter.get_name()
        self._storage_service = self.adapter.get_storage_service()

    def get_input_path(self, execution: Execution) -> str:
        return f"{self.adapter.staging_mount_path(execution)}/{execution.input_relative_path}"

    def get_output_path(self, execution: Execution) -> str:
        return f"{self.adapter.staging_mount_path(execution)}/{execution.output_relative_path}"

    def update_execution_env(self, execution: Execution):
        execution.add_envs["INPUT_PATH"] = self.get_input_path(execution)
        execution.add_envs["OUTPUT_PATH"] = self.get_output_path(execution)
        execution.add_envs["EXECUTION_ID"] = execution.id

    async def upload_python_script_if_needed_and_update_overridden_args(
        self, execution: Execution
    ):
        job = execution.job
        if not job.python_script_path:
            return
        with open(job.python_script_path) as f:
            content = f.read()
        destination_path = f"{execution.base_dir}/input/script.py"
        logger.info(
            "Uploading python script (%s) to %s/%s...",
            job.python_script_path,
            execution.config._staging_bucket,
            destination_path,
        )
        await self.storage_service.upload(
            content.encode("utf8"),
            execution.config._staging_bucket_name,
            destination_path,
        )
        logger.debug(
            "Done uploading python script (%s) to %s/%s",
            job.python_script_path,
            execution.config.staging_bucket,
            destination_path,
        )
        execution.overridden_args = [
            "python",
            f"{self.adapter.staging_mount_path(execution)}/{destination_path}",
        ]

    def replace_overridden_args_placeholders(self, execution: Execution):
        input_path = (
            f"{self.adapter.staging_mount_path(execution)}/{execution.base_dir}/input"
        )
        execution.overridden_args = [
            x.replace("{{INPUT}}", input_path) for x in execution.overridden_args
        ]

    async def upload_inputs(self, execution: Execution):
        async def _upload_input(input: Input):
            path = f"{execution.config._staging_bucket}/{execution.input_relative_path}/{input.filename}"
            logger.info(f"Uploading input to {path}...")
            await input._create(
                execution.config._staging_bucket_name,
                execution.input_relative_path,
                self.storage_service,
            )
            logger.debug(f"Done uploading input: {path}")

        inputs = execution.inputs
        await asyncio.gather(*[_upload_input(input) for input in inputs])

    async def create_input_output_paths_if_needed(self, execution: Execution):
        coroutines: list[Coroutine] = []
        logger.info(
            "Creating input path %s/%s/...",
            execution.config._staging_bucket,
            execution.input_relative_path,
        )
        coroutines.append(
            self.storage_service.upload(
                b"",
                execution.config._staging_bucket_name,
                execution.input_relative_path + "/",
            )
        )
        logger.info(
            "Creating output path %s/%s/...",
            execution.config._staging_bucket,
            execution.output_relative_path,
        )
        coroutines.append(
            self.storage_service.upload(
                b"",
                execution.config._staging_bucket_name,
                execution.output_relative_path + "/",
            )
        )
        await asyncio.gather(*coroutines, return_exceptions=True)
        logger.debug("Done creating input/output paths")

    async def prepare(self, execution: Execution):
        self.update_execution_env(execution)
        await self.create_input_output_paths_if_needed(execution)
        await self.upload_python_script_if_needed_and_update_overridden_args(execution)
        self.replace_overridden_args_placeholders(execution)
        await self.upload_inputs(execution)

    async def _schedule(
        self,
        job: SmartJob,
        execution_config: ExecutionConfig,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
    ) -> ExecutionResultFuture:
        execution = Execution(
            job,
            overridden_args=list(job.overridden_args),
            add_envs={**job.add_envs, **(add_envs or {})},
            config=execution_config,
            inputs=inputs or [],
        )
        with LogContext.bind(execution_id=execution.id):
            logger.info("Preparing the smartjob execution...")
            await self.prepare(execution)
            logger.debug("Smartjob execution prepared")
            logger.info("Scheduling the smartjob execution...")
            res = await self.adapter.schedule(execution)
            logger.debug("Smartjob execution scheduled")
            return res

    async def schedule(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
        execution_config: ExecutionConfig | None = None,
    ) -> ExecutionResultFuture:
        """Schedule a job and return a kind of future (coroutine version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.
            execution_config: Execution configuration.

        Returns:
            The result of the job execution as a kind of future.

        """
        execution_config = execution_config or ExecutionConfig()
        execution_config.fix_for_executor_name(self.executor_name)
        try:
            async with asyncio.timeout(
                execution_config._timeout_config._timeout_seconds_schedule
            ):
                with LogContext.bind(
                    job_name=job.name,
                    job_namespace=job.namespace,
                ):
                    try:
                        async for attempt in AsyncRetrying(
                            reraise=True,
                            wait=wait_exponential(multiplier=1, min=1, max=300),
                            stop=stop_after_attempt(
                                execution_config._retry_config._max_attempts_schedule
                            ),
                        ):
                            with attempt:
                                with LogContext.bind(
                                    attempt=attempt.retry_state.attempt_number
                                ):
                                    fut = await self._schedule(
                                        job,
                                        execution_config=execution_config,
                                        add_envs=add_envs,
                                        inputs=inputs,
                                    )
                                    logger.info(
                                        "Smartjob execution scheduled",
                                        log_url=fut.log_url,
                                    )
                                    LogContext.remove("attempt")
                                    LogContext.add(execution_id=fut.execution_id)
                                    fut._log_context = LogContext.getall()
                                    return fut
                    except RetryError:
                        raise
                    raise Exception("Unreachable code")
        except asyncio.TimeoutError:
            raise SmartJobTimeoutException(
                f"Timeout reached (after {execution_config._timeout_config._timeout_seconds_schedule} seconds)"
            )

    async def run(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
        execution_config: ExecutionConfig | None = None,
    ) -> ExecutionResult:
        """Schedule a job and wait for its completion (couroutine version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.
            execution_config: Execution configuration.

        Returns:
            The result of the job execution.

        """
        execution_config = execution_config or ExecutionConfig()
        execution_config.fix_timeout_config()
        try:
            async with asyncio.timeout(
                execution_config._timeout_config.timeout_seconds
            ):
                future = await self.schedule(
                    job,
                    add_envs=add_envs,
                    inputs=inputs,
                    execution_config=execution_config,
                )
                return await future.result()
        except asyncio.TimeoutError:
            raise SmartJobTimeoutException(
                f"Timeout reached ({execution_config._timeout_config.timeout_seconds} seconds)"
            )

    def sync_run(
        self,
        job: SmartJob,
        add_envs: dict[str, str] | None = None,
        inputs: list[Input] | None = None,
        execution_config: ExecutionConfig | None = None,
    ) -> ExecutionResult:
        """Schedule a job and wait for its completion (blocking version).

        Arguments:
            job: The job to run.
            add_envs: Environment variables to add for this particular execution.
            inputs: Inputs to add for this particular execution.
            execution_config: Execution configuration.

        Returns:
            The result of the job execution.
        """
        return asyncio.run(
            self.run(job, add_envs, inputs, execution_config=execution_config)
        )



import asyncio
import json
from dataclasses import dataclass, field
from typing import cast

import google.api_core.exceptions as gac_exceptions
from google.cloud import run_v2
from google.protobuf.duration_pb2 import Duration
from grpc import StatusCode
from stlog import LogContext, getLogger

from smartjob.app.execution import Execution, ExecutionResult
from smartjob.app.executor import ExecutionResultFuture
from smartjob.app.utils import hex_hash
from smartjob.infra.adapters.executor.gcp import GCPExecutionResultFuture, GCPExecutor

logger = getLogger("smartjob.executor.cloudrun")


@dataclass(frozen=True)
class JobListCacheKey:
    project: str
    region: str


class CloudRunExecutionResultFuture(GCPExecutionResultFuture):
    def _get_result_from_future(self, future: asyncio.Future) -> ExecutionResult:
        try:
            task_result = future.result()
            success = False
            castedResult = cast(run_v2.Execution, task_result)
            success = castedResult.succeeded_count > 0
            return ExecutionResult._from_execution(
                self._execution, success, self.log_url
            )
        except gac_exceptions.GoogleAPICallError:
            pass
        except asyncio.CancelledError:
            self._execution.cancelled = True
        return ExecutionResult._from_execution(self._execution, False, self.log_url)


@dataclass
class CloudRunExecutorAdapter(GCPExecutor):
    client_cache: run_v2.JobsAsyncClient | None = field(default=None, init=False)

    def job_hash(self, execution: Execution) -> str:
        job = execution.job
        conf = execution.config
        args: list[str] = [job.name, conf._project, conf._region, job.docker_image]
        args += [
            job.namespace,
            conf._staging_bucket,
            conf.service_account or "",
            json.dumps(conf.labels, sort_keys=True) if conf.labels else "",
        ]
        return hex_hash(*args)[:10]

    def parent_name(self, execution: Execution) -> str:
        conf = execution.config
        return f"projects/{conf._project}/locations/{conf._region}"

    def cloud_run_job_name(self, execution: Execution) -> str:
        return f"{execution.job.full_name}-{self.job_hash(execution)}"

    def full_cloud_run_job_name(self, execution: Execution) -> str:
        return (
            f"{self.parent_name(execution)}/jobs/{self.cloud_run_job_name(execution)}"
        )

    @property
    def client(self) -> run_v2.JobsAsyncClient:
        # lazy-init to circumvent some 'attached to a different loop' issues
        if self.client_cache is None:
            self.client_cache = run_v2.JobsAsyncClient()
        return self.client_cache

    async def _create_job_if_needed(self, execution: Execution):
        job = execution.job
        config = execution.config
        volumes: list[run_v2.Volume] = []
        volume_mounts: list[run_v2.VolumeMount] = []
        launch_stage = "GA"
        volumes.append(
            run_v2.Volume(
                name="staging",
                gcs=run_v2.GCSVolumeSource(
                    bucket=config._staging_bucket_name, read_only=False
                ),
            )
        )
        volume_mounts.append(run_v2.VolumeMount(name="staging", mount_path="/staging"))
        launch_stage = "BETA"
        vpc_access: run_v2.VpcAccess | None = None
        if config.vpc_connector_network and config.vpc_connector_subnetwork:
            vpc_access = run_v2.VpcAccess(
                egress=run_v2.VpcAccess.VpcEgress.PRIVATE_RANGES_ONLY,
                network_interfaces=[
                    run_v2.VpcAccess.NetworkInterface(
                        network=config.vpc_connector_network,
                        subnetwork=config.vpc_connector_subnetwork,
                    )
                ],
            )

        request = run_v2.CreateJobRequest(
            parent=self.parent_name(execution),
            job_id=self.cloud_run_job_name(execution),
            job=run_v2.Job(
                labels=execution.labels,
                launch_stage=launch_stage,
                template=run_v2.ExecutionTemplate(
                    task_count=1,
                    template=run_v2.TaskTemplate(
                        max_retries=config._retry_config._max_attempts_execute - 1,
                        execution_environment=run_v2.ExecutionEnvironment(
                            run_v2.ExecutionEnvironment.EXECUTION_ENVIRONMENT_GEN2
                        ),
                        containers=[
                            run_v2.Container(
                                name="container-1",
                                image=job.docker_image,
                                resources=run_v2.ResourceRequirements(
                                    startup_cpu_boost=False,
                                    limits={
                                        "cpu": f"{config._cpu}",
                                        "memory": f"{config._memory_gb}Gi",
                                    },
                                ),
                                volume_mounts=volume_mounts,
                            )
                        ],
                        volumes=volumes,
                        service_account=config.service_account,
                        vpc_access=vpc_access,
                    ),
                ),
            ),
        )
        logger.info("Let's create a new Cloud Run Job...")
        operation = await self.client.create_job(request=request)
        await operation.result()
        logger.debug("Done creating Cloud Run Job")

    async def create_job_if_needed(self, execution: Execution):
        try:
            await self._create_job_if_needed(execution)
        except gac_exceptions.GoogleAPICallError as e:
            if e.grpc_status_code != StatusCode.ALREADY_EXISTS:
                raise
            # we can ignore this
            logger.debug("Done creating Cloud Run Job (already exists)")

    async def schedule(self, execution: Execution) -> ExecutionResultFuture:
        job = execution.job
        config = execution.config
        job_id = self.cloud_run_job_name(execution)
        full_name = self.full_cloud_run_job_name(execution)
        with LogContext.bind(
            cloudrun_job_id=job_id, project=config._project, region=config._region
        ):
            await self.create_job_if_needed(execution)
            request = run_v2.RunJobRequest(
                name=full_name,
                overrides=run_v2.RunJobRequest.Overrides(
                    task_count=1,
                    timeout=Duration(seconds=config._timeout_config.timeout_seconds),
                    container_overrides=[
                        run_v2.RunJobRequest.Overrides.ContainerOverride(
                            name="container-1",
                            args=execution.overridden_args,
                            env=[
                                run_v2.EnvVar(name=k, value=v)
                                for k, v in execution.add_envs.items()
                            ],
                        )
                    ],
                ),
            )
            logger.info(
                "Let's trigger a new execution of Cloud Run Job...",
                docker_image=job.docker_image,
                overridden_args=execution.overridden_args_as_string,
                add_envs=execution.add_envs_as_string,
                timeout_s=config._timeout_config.timeout_seconds,
            )
            operation = await self.client.run_job(request=request)
            return CloudRunExecutionResultFuture(
                asyncio.create_task(operation.result()),
                execution=execution,
                storage_service=self.storage_service,
                log_url=operation.metadata.log_uri,
            )

    def get_name(self) -> str:
        return "cloudrun"

    def staging_mount_path(self, execution: Execution) -> str:
        return "/staging"



import typer
from stlog import setup

from smartjob.app.input import GcsInput, Input, LocalPathInput
from smartjob.app.job import DEFAULT_NAMESPACE


def add_env_argument_to_dict(add_env: list[str]) -> dict[str, str]:
    add_envs: dict[str, str] = {}
    for ae in add_env:
        if "=" not in ae:
            raise ValueError(f"Invalid env var format: {ae} => must be key=value")
        key, value = ae.split("=", 1)
        add_envs[key] = value
    return add_envs


def local_path_input_to_list(local_path_input: list[str]) -> list[Input]:
    res: list[Input] = []
    for lpi in local_path_input:
        if "=" not in lpi:
            raise ValueError(
                f"Invalid input format: {lpi} => must be filename=local_path"
            )
        filename, local_path = lpi.split("=", 1)
        res.append(
            LocalPathInput(
                filename=filename,
                local_path=local_path,
            )
        )
    return res


def gcs_input_to_list(gcs_input: list[str]) -> list[Input]:
    res: list[Input] = []
    for lpi in gcs_input:
        if "=" not in lpi:
            raise ValueError(
                f"Invalid input format: {lpi} => must be filename=full_gcs_path"
            )
        filename, gcs_path = lpi.split("=", 1)
        if not gcs_path.startswith("gs://"):
            raise ValueError(f"Invalid GCS path: {gcs_path} => must start with gs://")
        if "/" not in gcs_path[5:]:
            raise ValueError(
                f"Invalid GCS path: {gcs_path} => must have a / after gs://"
            )
        res.append(
            GcsInput(
                filename=filename,
                gcs_path=gcs_path,
            )
        )
    return res


def init_stlog(log_level: str):
    setup(level=log_level)


CpuOption = typer.Option(None, help="Number of CPUs (cloudrun only)")
MemoryOption = typer.Option(None, help="Memory in Gb (cloudrun only)")
NamespaceArgument = typer.Option(
    DEFAULT_NAMESPACE, envvar="SMARTJOB_NAMESPACE", help="namespace (simple string)"
)
NameArgument = typer.Argument(help="Name of the job")
DockerImageArgument = typer.Option("", help="Docker image to use")
OverrideCommandArgument = typer.Option(
    "",
    help="Override docker image command and arguments",
)
AddEnvArgument = typer.Option(
    [],
    help="Add a new env var in the container environment (format: key=value, can be used multiple times)",
)
StagingBucketArgument = typer.Option(
    None,
    envvar="SMARTJOB_STAGING_BUCKET",
    help="staging bucket (starting with gs://) for loading python_script_path (cloudrun or vertex only)",
)
InputBucketBasePathArgument = typer.Option(
    None, envvar="SMARTJOB_INPUT_BUCKET_BASE_PATH"
)
OutputBucketBasePathArgument: str = typer.Option(
    None, envvar="SMARTJOB_OUTPUT_BUCKET_BASE_PATH"
)
InputBucketPathArgument = typer.Option("")
OutputBucketPathArgument = typer.Option("")
PythonScriptPathArgument = typer.Option(
    "", help="local path to python script to execute in the container"
)
LocalPathInputArgument = typer.Option(
    [],
    help="Local path input (format: filename=local_path, can be used multiple times)",
)
GcsPathInputArguments = typer.Option(
    [],
    help="GCS path input (format: filename=full_source_gcs_path, can be used multiple times)",
)
LogLevelArgument = typer.Option("INFO", envvar="LOG_LEVEL", help="log level")
ProjectArgument = typer.Option(None, envvar="SMARTJOB_PROJECT", help="GCP project")
RegionArgument = typer.Option(None, envvar="SMARTJOB_REGION", help="GCP region")
ExecutorArgument = typer.Option(
    "cloudrun", help="executor type ('vertex', 'cloudrun' or 'docker')"
)
ServiceAccountArgument = typer.Option(
    None, envvar="SMARTJOB_SERVICE_ACCOUNT", help="GCP service account to use"
)
MaxAttemptsArgument = typer.Option(
    3, help="Number of attempts (if failure), must be >= 1"
)
VpcConnectorNetwork = typer.Option(None, help="VPC connector network (cloudrun only)")
VpcConnectorSubNetwork = typer.Option(
    None, help="VPC connector subnetwork (cloudrun only)"
)



class SmartJobException(Exception):
    """Base class for exceptions in this module."""

    pass


class SmartJobTimeoutException(SmartJobException):
    """Exception raised when a job execution times out."""

    pass



import asyncio
import json
import shlex
import sys

import typer
import typer.core

from smartjob.app.execution import ExecutionConfig
from smartjob.app.executor import ExecutionResultFuture
from smartjob.app.job import SmartJob
from smartjob.app.retry import RetryConfig
from smartjob.infra.controllers.cli.common import (
    AddEnvArgument,
    CpuOption,
    DockerImageArgument,
    ExecutorArgument,
    GcsPathInputArguments,
    LocalPathInputArgument,
    LogLevelArgument,
    MaxAttemptsArgument,
    MemoryOption,
    NameArgument,
    NamespaceArgument,
    OverrideCommandArgument,
    ProjectArgument,
    PythonScriptPathArgument,
    RegionArgument,
    ServiceAccountArgument,
    StagingBucketArgument,
    VpcConnectorNetwork,
    VpcConnectorSubNetwork,
    add_env_argument_to_dict,
    gcs_input_to_list,
    init_stlog,
    local_path_input_to_list,
)
from smartjob.infra.controllers.lib import get_executor_service

# Disable rich usage for CLI doc generation
typer.core.rich = None  # type: ignore

cli = typer.Typer(add_completion=False)


if __name__ == "__main__":
    cli()


@cli.command()
def run(
    name: str = NameArgument,
    docker_image: str = DockerImageArgument,
    override_command_and_args: str = OverrideCommandArgument,
    add_env: list[str] = AddEnvArgument,
    staging_bucket: str = StagingBucketArgument,
    python_script_path: str = PythonScriptPathArgument,
    cpu: float = CpuOption,
    memory_gb: float = MemoryOption,
    local_path_input: list[str] = LocalPathInputArgument,
    gcs_input: list[str] = GcsPathInputArguments,
    log_level: str = LogLevelArgument,
    namespace: str = NamespaceArgument,
    project: str = ProjectArgument,
    region: str = RegionArgument,
    executor: str = ExecutorArgument,
    service_account: str = ServiceAccountArgument,
    max_attempts: int = MaxAttemptsArgument,
    vpc_connector_network: str = VpcConnectorNetwork,
    vpc_connector_subnetwork: str = VpcConnectorSubNetwork,
):
    init_stlog(log_level)
    executor_service = get_executor_service(executor)
    add_envs = add_env_argument_to_dict(add_env)
    job = SmartJob(
        name=name,
        namespace=namespace,
        docker_image=docker_image,
        add_envs=add_envs,
        python_script_path=python_script_path,
        overridden_args=shlex.split(override_command_and_args),
    )
    execution_config = ExecutionConfig(
        project=project,
        region=region,
        staging_bucket=staging_bucket,
        cpu=cpu,
        memory_gb=memory_gb,
        service_account=service_account,
        retry_config=RetryConfig(max_attempts=max_attempts),
        vpc_connector_network=vpc_connector_network,
        vpc_connector_subnetwork=vpc_connector_subnetwork,
    )
    inputs = local_path_input_to_list(local_path_input) + gcs_input_to_list(gcs_input)
    result = executor_service.sync_run(
        job, inputs=inputs, execution_config=execution_config
    )
    return_code = 0
    if result:
        print("SUCCESS in %i seconds" % result.duration_seconds)
    else:
        print("FAILED in %i seconds" % result.duration_seconds)
        return_code = 2
    print("Id:      %s" % result.execution_id)
    print("Logs:    %s" % result.log_url)
    if result.json_output is not None:
        print("JsonOutput:")
        print(json.dumps(result.json_output, indent=4))
    sys.exit(return_code)


async def _schedule(
    executor_service, job, inputs, execution_config
) -> ExecutionResultFuture:
    return await executor_service.schedule(
        job, inputs=inputs, execution_config=execution_config
    )


@cli.command()
def schedule(
    name: str = NameArgument,
    docker_image: str = DockerImageArgument,
    override_command_and_args: str = OverrideCommandArgument,
    add_env: list[str] = AddEnvArgument,
    staging_bucket: str = StagingBucketArgument,
    python_script_path: str = PythonScriptPathArgument,
    cpu: float = typer.Option(1.0, help="Number of CPUs"),
    memory_gb: float = typer.Option(0.5, help="Memory in Gb"),
    local_path_input: list[str] = LocalPathInputArgument,
    gcs_input: list[str] = GcsPathInputArguments,
    log_level: str = LogLevelArgument,
    namespace: str = NamespaceArgument,
    project: str = ProjectArgument,
    region: str = RegionArgument,
    executor: str = ExecutorArgument,
    service_account: str = ServiceAccountArgument,
    max_attempts: int = MaxAttemptsArgument,
    vpc_connector_network: str = VpcConnectorNetwork,
    vpc_connector_subnetwork: str = VpcConnectorSubNetwork,
):
    init_stlog(log_level)
    executor_service = get_executor_service(executor)
    add_envs = add_env_argument_to_dict(add_env)
    job = SmartJob(
        name=name,
        namespace=namespace,
        docker_image=docker_image,
        add_envs=add_envs,
        python_script_path=python_script_path,
        overridden_args=shlex.split(override_command_and_args),
    )
    execution_config = ExecutionConfig(
        project=project,
        region=region,
        staging_bucket=staging_bucket,
        cpu=cpu,
        memory_gb=memory_gb,
        service_account=service_account,
        retry_config=RetryConfig(max_attempts=max_attempts),
        vpc_connector_network=vpc_connector_network,
        vpc_connector_subnetwork=vpc_connector_subnetwork,
    )
    inputs = local_path_input_to_list(local_path_input) + gcs_input_to_list(gcs_input)
    result_future = asyncio.run(
        _schedule(executor_service, job, inputs, execution_config)
    )
    print("SCHEDULED!")
    print("Id:      %s" % result_future.execution_id)
    print("Logs:    %s" % result_future.log_url)
    result_future._cancel()



from dataclasses import dataclass


@dataclass
class TimeoutConfig:
    """Timeout configuration."""

    timeout_seconds: int = 3600
    """FIXME"""

    timeout_seconds_schedule: int | None = None
    """FIXME"""

    timeout_seconds_download: int | None = None
    """FIXME"""

    @property
    def _timeout_seconds_schedule(self) -> int:
        return self.timeout_seconds_schedule or self.timeout_seconds

    @property
    def _timeout_seconds_download(self) -> int:
        return self.timeout_seconds_download or self.timeout_seconds

    def __post_init__(self):
        for field in (
            "timeout_seconds",
            "timeout_seconds_schedule",
            "timeout_seconds_download",
        ):
            if getattr(self, field) is not None and getattr(self, field) <= 0:
                raise ValueError(f"{field} must be > 0")
